traceabilityReport = new TraceabilityReportTool(maxDistance: maxDistance, excludedLinkTypeIds: excludedLinkTypeIds, excludedIssueTypeIds: excludedIssueTypeIds, traceOutgoingLinks: traceOutgoingLinks, traceIncomingLinks: traceIncomingLinks, traceSubTasks: traceSubTasks)

class TraceabilityReportTool {
	def maxDistance
	def excludedLinkTypeIds
	def excludedIssueTypeIds
	def traceOutgoingLinks
	def traceIncomingLinks
	def traceSubTasks

	/**
	 * Returns a list of traceability items (an issue with the link pointing to it).
	 */
	public getTraceabilityReportItemsFrom(issues) {
		def items = []

		for (issue in issues) {
			traverseIssueLinks(items, 0, issue, null, null)
		}

		return items
	}

	/**
	 * @param fromIssue the issue from which the traveral arrived to the current issue, or null for top-level issues (sources of traceability).
	 */
	private void traverseIssueLinks(items, currentDistance, issue, fromIssue, linkName) {
		// add current issue
		items.add(new TraceabilityItem(distance: currentDistance, linkName: linkName, issue: issue))

		// traverse relations
		if (currentDistance < maxDistance) {
			// traverse links
			for (link in issue.issueLinks) {
				boolean isOutwardLink = (link.outwardIssue != null)
				def linkedIssue = isOutwardLink ? link.outwardIssue : link.inwardIssue

				boolean isTraversableLink = (linkedIssue != fromIssue) // don't traverse backwards
				boolean isExcludedLink = excludedLinkTypeIds.contains(link.type.id as Integer) ||
										 excludedIssueTypeIds.contains(link.outwardIssue?.issueType?.id as Integer) ||
										 excludedIssueTypeIds.contains(link.inwardIssue?.issueType?.id as Integer) ||
										 (traceOutgoingLinks == false && isOutwardLink) ||
										 (traceIncomingLinks == false && !isOutwardLink)
				if (isTraversableLink && !isExcludedLink) {
					traverseIssueLinks(items, currentDistance + 1, linkedIssue, issue, isOutwardLink ? link.type.outward : link.type.inward)
				}
			}

			// traverse sub-tasks
			if (traceSubTasks) {
				for (subTask in issue.subTaskObjects) {
					traverseIssueLinks(items, currentDistance + 1, subTask, issue, "has sub-task")
				}
			}
		}
	}
}

class TraceabilityItem {
	/**
	 * 1 means directly linked from a source issue.
	 * 2 means linked from an issue that is directly linked from a source issue.
	 */
	def distance;
	def linkName;
	def issue;
}
